"""This module contains the DirectBoxSizer class."""

__all__ = ['DirectBoxSizer']

from panda3d.core import *
from direct.gui import DirectGuiGlobals as DGG
from direct.gui.DirectFrame import DirectFrame
from . import DirectGuiHelper as DGH


DGG.HORIZONTAL_INVERTED = 'horizontal_inverted'


class DirectItemContainer():
    def __init__(self, element, **kw):
        self.element = element
        #if "stretch" in kw:
        #    if kw.get("stretch"):
        #        pass
        #    del kw["stretch"]
        #self.margin = VBase4(0,0,0,0)
        #if "margin" in kw:
        #    self.margin = kw.get("margin")

class DirectBoxSizer(DirectFrame):
    """
    A frame to add multiple other directGui elements to that will then be
    automatically be placed stacked next to each other.
    """

    # Horizontal
    A_Center = 0b1
    A_Left = 0b10
    A_Right = 0b100

    # Vertical
    A_Middle = 0b1000
    A_Top = 0b10000
    A_Bottom = 0b100000

    def __init__(self, parent = None, **kw):
        optiondefs = (
            # Define type of DirectGuiWidget
            ('items',          [],          self.refresh),
            ('pgFunc',         PGItem,      None),
            ('numStates',      1,           None),
            ('state',          DGG.NORMAL,  None),
            ('borderWidth',    (0, 0),      self.setBorderWidth),

            ('orientation', DGG.HORIZONTAL, self.refresh),
            ('itemMargin',     (0,0,0,0),   self.refresh),
            ('itemAlign',  self.A_Left|self.A_Top, self.refresh),
            ('autoUpdateFrameSize', True,   None),

            ('suppressMouse',  0,           None),
            )
        # Merge keyword options with default options
        self.defineoptions(kw, optiondefs)

        # Initialize superclasses
        DirectFrame.__init__(self, parent)

        # Call option initialization functions
        self.initialiseoptions(DirectBoxSizer)

        self.itemsLeft = 0
        self.itemsRight = 0
        self.itemsBottom = 0
        self.itemsTop = 0

    def addItem(self, element, **kw):
        """
        Adds the given item to this panel stack
        """
        element.reparentTo(self)
        container = DirectItemContainer(element, **kw)
        self["items"].append(container)
        self.refresh()

    def removeItem(self, element):
        """
        Remove this item from the panel
        """
        for item in self["items"]:
            if element == item.element:
                self["items"].remove(item)
                self.refresh()
                return 1
        return 0

    def refresh(self):
        """
        Recalculate the position of every item in this panel and set the frame-
        size of the panel accordingly if auto update is enabled.
        """
        # sanity check so we don't get here to early
        if not hasattr(self, "bounds"): return
        if len(self["items"]) == 0: return


        for item in self["items"]:
            item.element.frameInitialiseFunc()

        nextX = 0
        nextY = 0

        sizer_pad = self["pad"]

        b_top = 0
        b_bottom = 0
        b_left = 0
        b_right = 0

        margin_left = self["itemMargin"][0]
        margin_right = self["itemMargin"][1]
        margin_bottom = self["itemMargin"][2]
        margin_top = self["itemMargin"][3]

        if self["autoUpdateFrameSize"]:
            # set the frame size to 0, we'll calculate the new size at the end
            self["frameSize"] = (0, 0, 0, 0)

        itemsWidth = 0
        itemsHeight = 0
        if self['orientation'] == DGG.HORIZONTAL:
            # Horizontal - Left to Right placement
            if not self["autoUpdateFrameSize"]:
                # get the left side of the box sizer frame
                if self["frameSize"] != None:
                    nextX = self["frameSize"][0]

            # go through all items in the box and place them
            for item in self["items"]:
                curElem = item.element
                y = -DGH.getRealBottom(curElem)

                # Vertical Alingment
                if self["itemAlign"] & self.A_Middle:
                    # Items are alligned by their center
                    y -= DGH.getRealHeight(curElem) / 2

                    # get the new top and bottom if we should resize the box
                    b_bottom = min(b_bottom, -DGH.getRealHeight(curElem) / 2)
                    b_top = max(b_top, DGH.getRealHeight(curElem) / 2)
                elif self["itemAlign"] & self.A_Top:
                    # Items are alligned by their upper edge
                    y -= DGH.getRealHeight(curElem)
                    # Vertical adjustment to the box' size
                    y += self["frameSize"][3]

                    # get the new top and bottom if we should resize the box
                    b_bottom = min(b_bottom, -DGH.getRealHeight(curElem))
                    b_top = 0
                else:
                    # Items are alligned by their lower edge
                    b_top = max(b_top, DGH.getRealHeight(curElem))
                    # Vertical adjustment to the box' size
                    y += self["frameSize"][2]

                # place the element and calculate the next x position
                curElem.setPos(nextX - DGH.getRealLeft(curElem), 0, y)
                itemWidth = (DGH.getRealWidth(curElem) + margin_left + margin_right)
                nextX += itemWidth
                b_right = nextX
                itemsWidth += itemWidth

        elif self['orientation'] == DGG.HORIZONTAL_INVERTED:
            # Horizontal - Right to Left
            if not self["autoUpdateFrameSize"]:
                if self["frameSize"] != None:
                    nextX = self["frameSize"][1]
            for item in self["items"]:
                curElem = item.element
                y = -DGH.getRealBottom(curElem)

                # Vertical Alingment
                if self["itemAlign"] & self.A_Middle:
                    # Items are alligned by their center
                    y -= DGH.getRealHeight(curElem) / 2

                    # get the new top and bottom if we should resize the box
                    b_bottom = min(b_bottom, -DGH.getRealHeight(curElem) / 2)
                    b_top = max(b_top, DGH.getRealHeight(curElem) / 2)
                elif self["itemAlign"] & self.A_Top:
                    # Items are alligned by their upper edge
                    y -= DGH.getRealHeight(curElem)
                    # Vertical adjustment to the box' size
                    y += self["frameSize"][3]

                    # get the new top and bottom if we should resize the box
                    b_bottom = min(b_bottom, -DGH.getRealHeight(curElem))
                    b_top = 0
                else:
                    # Items are alligned by their lower edge
                    b_top = max(b_top, DGH.getRealHeight(curElem))
                    # Vertical adjustment to the box' size
                    y += self["frameSize"][2]

                # place the element and calculate the next x position
                curElem.setPos(nextX + DGH.getRealRight(curElem), 0, y)
                itemWidth = (DGH.getRealWidth(curElem) + margin_left + margin_right)
                nextX -= itemWidth
                b_left = nextX
                itemsWidth += itemWidth

        elif self['orientation'] == DGG.VERTICAL:
            # Vertical - Top to Bottom
            if not self["autoUpdateFrameSize"]:
                if self["frameSize"] != None:
                    nextY = self["frameSize"][3]
            for item in self["items"]:
                curElem = item.element
                x = sizer_pad[0]

                # Horizontal Alingment
                if self["itemAlign"] & self.A_Left:
                    # Items are alligned by their Left edge
                    x -= DGH.getRealLeft(curElem)
                    # Horizontal adjustment to the box' size
                    x += self["frameSize"][0]

                    # get the new left and right if we should resize the box
                    b_left = 0
                    b_right = max(b_right, DGH.getRealWidth(curElem))
                elif self["itemAlign"] & self.A_Right:
                    # Items are alligned by their Right edge
                    x -= DGH.getRealRight(curElem)
                    # Horizontal adjustment to the box' size
                    x += self["frameSize"][1]

                    # get the new left and right if we should resize the box
                    b_left = min(b_left, -DGH.getRealWidth(curElem))
                    b_right = 0
                else:
                    # Items are alligned by their Center
                    b_left = min(b_left, DGH.getRealLeft(curElem))
                    b_right = max(b_right, DGH.getRealRight(curElem))

                # place the element and calculate the next y position
                curElem.setPos(x, 0, nextY - DGH.getRealTop(curElem))
                itemHeight = (DGH.getRealHeight(curElem) + margin_top + margin_bottom)
                nextY -= itemHeight
                b_bottom = nextY
                itemsHeight += itemHeight

        elif self['orientation'] == DGG.VERTICAL_INVERTED:
            # Vertical - Bottom to Top
            if not self["autoUpdateFrameSize"]:
                if self["frameSize"] != None:
                    nextY = self["frameSize"][2]
            for item in self["items"]:
                curElem = item.element
                x = sizer_pad[0]
                if self["itemAlign"] & self.A_Left:
                    # Items are alligned by their Left edge
                    x -= DGH.getRealLeft(curElem)
                    # Horizontal adjustment to the box' size
                    x += self["frameSize"][0]

                    # get the new left and right if we should resize the box
                    b_left = 0
                    b_right = max(b_right, DGH.getRealWidth(curElem))
                elif self["itemAlign"] & self.A_Right:
                    x -= DGH.getRealRight(curElem)
                    # Horizontal adjustment to the box' size
                    x += self["frameSize"][1]

                    # get the new left and right if we should resize the box
                    b_left = min(b_left, -DGH.getRealWidth(curElem))
                    b_right = 0
                else:
                    # Items are alligned by their Center
                    b_left = min(b_left, DGH.getRealLeft(curElem))
                    b_right = max(b_right, DGH.getRealRight(curElem))

                # place the element and calculate the next y position
                curElem.setPos(x, 0, nextY - DGH.getRealBottom(curElem))
                itemHeight = (DGH.getRealHeight(curElem) + margin_top + margin_bottom)
                nextY += itemHeight
                b_bottom = nextY
                itemsHeight += itemHeight

        else:
            raise ValueError('Invalid value for orientation: %s' % (self['orientation']))

        if self["autoUpdateFrameSize"]:
            self["frameSize"] = (b_left+sizer_pad[0], b_right+sizer_pad[0], b_bottom+sizer_pad[1], b_top+sizer_pad[1])
        else:
            if self['orientation'] in [DGG.HORIZONTAL, DGG.HORIZONTAL_INVERTED]:
                # update Horizontal Align for the whole item block
                if self["itemAlign"] & self.A_Center:
                    if self['orientation'] == DGG.HORIZONTAL_INVERTED:
                        xShift = DGH.getRealWidth(self)/2 - itemsWidth/2
                    else:
                        xShift = DGH.getRealWidth(self)/2 - itemsWidth/2
                    for item in self["items"]:
                        if self['orientation'] == DGG.HORIZONTAL_INVERTED:
                            #item.element.setX(item.element.getX() - xShift)
                            pass
                        else:
                            item.element.setX(item.element.getX() + xShift)
                elif self["itemAlign"] & self.A_Right:
                    xShift = DGH.getRealWidth(self)
                    if self['orientation'] == DGG.HORIZONTAL_INVERTED:
                        xShift = -xShift
                    for item in self["items"]:
                        item.element.setX(item.element.getX() + xShift)

            elif self['orientation'] in [DGG.VERTICAL, DGG.VERTICAL_INVERTED]:
                # update Vertical Align for Horizontally elements
                if self["itemAlign"] & self.A_Middle:
                    if self['orientation'] == DGG.VERTICAL_INVERTED:
                        yShift = DGH.getRealHeight(self)/2 + itemsHeight/2
                    else:
                        yShift = DGH.getRealHeight(self)/2 - itemsHeight/2
                    for item in self["items"]:
                        item.element.setZ(item.element.getZ() - yShift)
                elif self["itemAlign"] & self.A_Bottom:
                    if self['orientation'] == DGG.VERTICAL_INVERTED:
                        yShift = DGH.getRealHeight(self)/2 + itemsHeight/2
                    else:
                        yShift = DGH.getRealHeight(self) - itemsHeight
                    for item in self["items"]:
                        item.element.setZ(item.element.getZ() - yShift)

        #TODO: Maybe the item index has to be changed for inverted positioning
        # store item block edges
        self.itemsLeft = b_left+sizer_pad[0] + self.getX()
        self.itemsRight = b_right+sizer_pad[0] + self.getX()
        self.itemsBottom = b_bottom+sizer_pad[1]# + self.getZ()
        self.itemsTop = b_top+sizer_pad[1]# + self.getZ()



#!/usr/bin/python
# -*- coding: utf-8 -*-
__author__ = "Fireclaw the Fox"
__license__ = """
Simplified BSD (BSD 2-Clause) License.
See License.txt or http://opensource.org/licenses/BSD-2-Clause for more info
"""

from panda3d import core as p3d


def getBorderSize(guiItem):
    frameType = guiItem.getFrameType()
    if guiItem["frameSize"] is not None \
    or frameType == p3d.PGFrameStyle.TNone \
    or frameType == p3d.PGFrameStyle.TFlat:
        return (0,0)
    return guiItem["borderWidth"]

def getBounds(guiItem):
    #HACK: Sometimes getBounds returns 0s while a frameSize is actually given
    if guiItem.getBounds() == [0,0,0,0] and guiItem['frameSize'] != (0,0,0,0):
        if guiItem['frameSize'] is not None:
            return guiItem['frameSize']
        else:
            if guiItem.guiItem.getFrame() is not None:
                return guiItem.guiItem.getFrame()
            # well... seems like this element just has no size.
            return guiItem.getBounds()
    return guiItem.getBounds()

def getRealWidth(guiItem):
    return (guiItem.getWidth() + 2 * getBorderSize(guiItem)[0]) * guiItem.getScale()[0]

def getRealHeight(guiItem):
    return (guiItem.getHeight() + 2 * getBorderSize(guiItem)[1]) * guiItem.getScale()[1]

def getRealLeft(guiItem):
    return (getBounds(guiItem)[0] - getBorderSize(guiItem)[0]) * guiItem.getScale()[0]

def getRealRight(guiItem):
    return (getBounds(guiItem)[1] + getBorderSize(guiItem)[0]) * guiItem.getScale()[0]

def getRealTop(guiItem):
    return (getBounds(guiItem)[3] + getBorderSize(guiItem)[1]) * guiItem.getScale()[1]

def getRealBottom(guiItem):
    return (getBounds(guiItem)[2] - getBorderSize(guiItem)[1]) * guiItem.getScale()[1]



"""Implements a pop-up menu containing multiple clickable options and sub-menus."""

__all__ = ['DirectMenuItem']

from panda3d.core import *
from direct.gui import DirectGuiGlobals as DGG
from direct.gui.DirectButton import *
from direct.gui.DirectLabel import *
from direct.gui.DirectFrame import *
from .DirectBoxSizer import DirectBoxSizer

DGG.MWUP = PGButton.getPressPrefix() + MouseButton.wheel_up().getName() + '-'
DGG.MWDOWN = PGButton.getPressPrefix() + MouseButton.wheel_down().getName() + '-'
DGG.LEFT = "left"
DGG.RIGHT = "right"
DGG.ABOVE = "above"
DGG.BELOW = "below"

class DirectMenuItemEntry:
    def __init__(self, text, command, extraArgs=None):
        self.text = text
        self.command = command
        self.extraArgs = extraArgs

class DirectMenuItemSubMenu:
    def __init__(self, text, items):
        self.text = text
        self.items = items

class DirectMenuItem(DirectButton):
    def __init__(self, parent = None, **kw):
        optiondefs = (
            # List of items to display on the popup menu
            ('items',       [],             self.setItems),
            # The position of the popup menu
            # possible positions: left, above, right, below
            ('popupMenuLocation', DGG.BELOW, None),
            # Background color to use to highlight popup menu items
            ('highlightColor', (.5, .5, .5, 1), None),
            # Background color of unhighlighted menu items
            ('itemFrameColor', None, None),
            # Alignment to use for text on popup menu button
            # Changing this breaks button layout
            ('text_align',  TextNode.ALeft, None),
            # Remove press effect because it looks a bit funny
            ('pressEffect',     0,          DGG.INITOPT),
            ('isSubMenu',       False,      None),
            ('parentMenu',      None,       None),
           )
        # Merge keyword options with default options
        self.defineoptions(kw, optiondefs)
        # Initialize superclasses
        DirectButton.__init__(self, parent)
        # This is created when you set the menu's items
        self.popupMenu = None
        # the selected item
        self.highlightedItem = None
        # A big screen encompassing frame to catch the cancel clicks
        self.cancelFrame = self.createcomponent(
            'cancelframe', (), None,
            DirectFrame, (self,),
            frameSize = (-1, 1, -1, 1),
            relief = None,
            state = 'normal')
        # Make sure this is on top of all the other widgets
        self.cancelFrame.setBin('gui-popup', 0)
        self.cancelFrame.node().setBounds(OmniBoundingVolume())
        self.cancelFrame.bind(DGG.B1PRESS, self.hidePopupMenu, extraArgs=[True])
        # Default action on press is to show popup menu
        self.bind(DGG.B1PRESS, self.showPopupMenu)
        # Check if item is highlighted on release and select it if it is
        self.bind(DGG.B1RELEASE, self._selectHighlighted)
        # Call option initialization functions
        self.initialiseoptions(DirectMenuItem)
        # Need to call this since we explicitly set frame size
        self.resetFrameSize()

    def setItems(self):
        """
        self['items'] = list of DirectMenuItemEntry and DirectMenuItemSubMenu
        Create new popup menu to reflect specified set of items
        """
        # Remove old component if it exits
        if self.popupMenu != None:
            self.destroycomponent('popupMenu')
        # Create new component
        self.popupMenu = self.createcomponent('popupMenu', (), None,
                                              DirectBoxSizer,
                                              (self,),
                                              itemAlign = TextNode.ALeft,
                                              orientation = DGG.VERTICAL,
                                              )
        # Make sure it is on top of all the other gui widgets
        self.popupMenu.setBin('gui-popup', 0)
        if not self['items']:
            return
        # Create a new component for each item
        # Find the maximum extents of all items
        itemIndex = 0
        self.minX = self.maxX = self.minZ = self.maxZ = None
        for item in self['items']:
            if type(item) is DirectMenuItemSubMenu:
                c = self.createcomponent(
                    'item%d' % itemIndex, (), 'item',
                    DirectMenuItem,
                    (self.popupMenu,),
                    text = item.text,
                    popupMenuLocation=DGG.RIGHT,
                    items=item.items,
                    item_relief = self["item_relief"],
                    isSubMenu=True,
                    parentMenu=self)
            else:
                c = self.createcomponent(
                    'item%d' % itemIndex, (), 'item',
                    DirectButton,
                    (self.popupMenu,),
                    text = item.text,
                    text_align = TextNode.ALeft,
                    command = item.command,
                    extraArgs = item.extraArgs,
                    frameColor = self["itemFrameColor"] if self["itemFrameColor"] else self["frameColor"])

                c.bind(DGG.B1RELEASE, self.hidePopupMenu, extraArgs=[True])
            bounds = c.getBounds()

            c.resetFrameSize()

            self.minX = min(self.minX if self.minX else bounds[0], bounds[0])
            self.maxX = max(self.maxX if self.maxX else bounds[1], bounds[1])
            self.minZ = min(self.minZ if self.minZ else bounds[2], bounds[2])
            self.maxZ = max(self.maxZ if self.maxZ else bounds[3], bounds[3])

            self.popupMenu.addItem(c)

            # Highlight background when mouse is in item
            c.bind(DGG.WITHIN,
                      lambda x, item=c: self._highlightItem(item))
            # Restore specified color upon exiting
            fc = self['itemFrameColor'] if self['itemFrameColor'] else c['frameColor']
            c.bind(DGG.WITHOUT,
                      lambda x, item=c, fc=fc: self._unhighlightItem(item, fc))
            c.bind(DGG.MWDOWN, self.scrollPopUpMenu, [-1])
            c.bind(DGG.MWUP, self.scrollPopUpMenu, [1])

            itemIndex += 1

        # Calc max width and height
        self.maxWidth = self.maxX - self.minX
        self.maxHeight = self.maxZ - self.minZ
        # Adjust frame size for each item and bind actions to mouse events
        for i in self.popupMenu["items"]:
            item = i.element

            # make all entries the same size
            item['frameSize'] = (self.minX, self.maxX, self.minZ, self.maxZ)


        # HACK: Set the user defined popup menu relief here so we don't
        # break the bounds calculation.
        self.popupMenu.setRelief(self['popupMenu_relief'])

        # Set initial state
        self.hidePopupMenu()
        self.bind(DGG.MWDOWN, self.scrollPopUpMenu, [-1])
        self.bind(DGG.MWUP, self.scrollPopUpMenu, [1])
        self.cancelFrame.bind(DGG.MWDOWN, self.scrollPopUpMenu, [-1])
        self.cancelFrame.bind(DGG.MWUP, self.scrollPopUpMenu, [1])

        self.popupMenu.refresh()

    def showPopupMenu(self, event = None):
        """
        Make popup visible.
        Adjust popup position if default position puts it outside of
        visible screen region
        """

        # Needed attributes (such as minZ) won't be set unless the user has specified
        # items to display. Let's assert that we've given items to work with.
        items = self['items']
        assert items and len(items) > 0, 'Cannot show an empty popup menu! You must add items!'

        # Show the menu
        self.popupMenu.show()
        # Compute bounds
        b = self.getBounds()
        fb = self.popupMenu.getBounds()
        self.popupMenu["itemAlign"] = TextNode.ALeft

        if self['popupMenuLocation'] == DGG.RIGHT:
            # This is the default to not break existing applications
            # Position menu at midpoint of button
            xPos = (b[1] - b[0]) - fb[0]
        elif self['popupMenuLocation'] == DGG.LEFT:
            # Position to the left
            xPos = b[0]
            self.popupMenu["itemAlign"] = TextNode.ARight
        else:
            # position to line up with the left edge if the menu is above or below
            xPos = b[0]
        self.popupMenu.setX(self, xPos)

        if self['popupMenuLocation'] == DGG.ABOVE:
            # Try to set height to line up selected item with button
            self.popupMenu.setZ(
                self, self.maxZ - fb[2])
        elif self['popupMenuLocation'] == DGG.BELOW:
            # Try to set height to line up selected item with button
            self.popupMenu.setZ(
                self, self.minZ)
        else:
            # Try to set height to line up selected item with button
            self.popupMenu.setZ(
                self, self.maxZ)
        # Make sure the whole popup menu is visible
        pos = self.popupMenu.getPos(render2d)
        scale = self.popupMenu.getScale(render2d)
        # How are we doing relative to the right side of the screen
        maxX = pos[0] + fb[1] * scale[0]
        if maxX > 1.0:
            # Need to move menu to the left
            self.popupMenu.setX(render2d, pos[0] + (1.0 - maxX))
        # How are we doing relative to the right side of the screen
        minX = pos[0]
        if minX < -1.0:
            # Need to move menu to the right
            self.popupMenu.setX(render2d, -1 )
        # How about up and down?
        minZ = pos[2] + fb[2] * scale[2]
        maxZ = pos[2] + fb[3] * scale[2]
        if minZ < -1.0:
            # Menu too low, move it up
            self.popupMenu.setZ(render2d, pos[2] + (-1.0 - minZ))

            # recheck the top position once repositioned
            pos = self.popupMenu.getPos(render2d)
            maxZ = pos[2] + fb[3] * scale[2]
            if maxZ > 1.0:
                # Menu too large to show on screen entirely
                # Try to set height to line up selected item with button
                self.popupMenu.setZ(
                    self, self.minZ)
        elif maxZ > 1.0:
            # Menu too high, move it down
            self.popupMenu.setZ(render2d, pos[2] + (1.0 - maxZ))
            # recheck the top position once repositioned
            pos = self.popupMenu.getPos(render2d)
            minZ = pos[2] + fb[2] * scale[2]
            if minZ < -1.0:
                # Menu too large to show on screen entirely
                # Try to set height to line up selected item with button
                self.popupMenu.setZ(
                    self, self.minZ)
        # Also display cancel frame to catch clicks outside of the popup
        self.cancelFrame.show()
        # Position and scale cancel frame to fill entire window
        self.cancelFrame.setPos(render2d, 0, 0, 0)
        self.cancelFrame.setScale(render2d, 1, 1, 1)

        self.popupMenu.refresh()

    def hidePopupMenu(self, event = None, hideParentMenu=False):
        """ Put away popup and cancel frame """
        self.popupMenu.hide()
        self.cancelFrame.hide()

        # call up the ancestry tree
        if hideParentMenu:
            if self['isSubMenu']:
                self['parentMenu'].hidePopupMenu(hideParentMenu=True)

    def scrollPopUpMenu(self, direction, event = None):
        """ Scroll the item frame up and down depending on the direction
        which must be a nummeric value. A positive value will scroll up
        while a negative value will scroll down. It will only work if
        items are out of bounds of the window """
        fb = self.popupMenu.getBounds()
        pos = self.popupMenu.getPos(render2d)
        scale = self.popupMenu.getScale(render2d)

        minZ = pos[2] + fb[2] * scale[2]
        maxZ = pos[2] + fb[3] * scale[2]
        if (minZ < -1.0 and direction > 0) or (maxZ > 1.0 and direction < 0):
            oldZ = self.popupMenu.getZ()
            #self.popupMenu.setZ(oldZ + direction * self.maxHeight)

    def _selectHighlighted(self, event=None):
        """
        Check to see if item is highlighted (by cursor being within
        that item).  If so, selected it.  If not, do nothing
        """
        if self.highlightedItem:
            # Pass any extra args to command
            self.highlightedItem['command'](*self.highlightedItem['extraArgs'])
            self.hidePopupMenu(hideParentMenu=True)

    def _highlightItem(self, item):
        """ Set frame color of highlighted item, record index """

        for i in self.popupMenu["items"]:
            # make sure all other items are unhighlighted
            base.messenger.send(DGG.WITHOUT + i.element.guiId, [""])

        item['frameColor'] = self['highlightColor']
        self.highlightedItem = item

        if type(item) is DirectMenuItem:
            if item['isSubMenu']:
                taskMgr.doMethodLater(0.2, item.showPopupMenu, "highlight")

    def _unhighlightItem(self, item, frameColor):
        """ Clear frame color """
        item['frameColor'] = frameColor
        self.highlightedItem = None

        if type(item) is DirectMenuItem:
            if item['isSubMenu']:
                def unhiglight(item):
                    if not item.highlightedItem:
                        item.hidePopupMenu()
                taskMgr.doMethodLater(0.2, unhiglight, "unhighlight", extraArgs=[item])



#Requires:
#    Can we get conjunction between two sizer to work to simultaneously size two splitframes at once?
#    Collapse function?





"""This module contains the DirectSplitFrame class."""

__all__ = ['DirectSplitFrame']

from panda3d.core import *
from direct.gui import DirectGuiGlobals as DGG
from direct.gui.DirectFrame import DirectFrame
from . import DirectGuiHelper as DGH


DGG.HORIZONTAL_INVERTED = 'horizontal_inverted'


class DirectItemContainer():
    def __init__(self, element, **kw):
        self.element = element
        #if "stretch" in kw:
        #    if kw.get("stretch"):
        #        pass
        #    del kw["stretch"]
        #self.margin = VBase4(0,0,0,0)
        #if "margin" in kw:
        #    self.margin = kw.get("margin")

class DirectSplitFrame(DirectFrame):
    """
    Two frames that can be resized with a small line between them
    """

    def __init__(self, parent = None, **kw):
        optiondefs = (
            # Define type of DirectGuiWidget
            ('pgFunc',         PGItem,      None),
            ('numStates',      1,           None),
            ('state',          DGG.NORMAL,  None),
            ('borderWidth',    (0, 0),      self.setBorderWidth),
            ('orientation', DGG.HORIZONTAL, self.refresh),
            ('framesize',      (-1,1,-1,1), None),

            # TODO: Change this. This only works for certain circumstances
            ('pixel2d',        False,       self.refresh),

            ('showSplitter',   True,        self.setSplitter),
            ('splitterPos',    0,           self.refresh),
            ('splitterWidth',  0.02,        self.refresh),
            ('splitterColor', (.7, .7, .7, 1), None),
            ('splitterHighlightColor', (.9, .9, .9, 1), None),

            ('firstFrameUpdateSizeFunc', None,  None),
            ('secondFrameUpdateSizeFunc', None, None),
            ('firstFrameMinSize', None,     None),
            ('secondFrameMinSize', None,    None),

            ('suppressMouse',  0,           None),
            )
        # Merge keyword options with default options
        self.defineoptions(kw, optiondefs)

        # Initialize superclasses
        DirectFrame.__init__(self, parent)

        # Call option initialization functions
        self.initialiseoptions(DirectSplitFrame)

        self.resetFrameSize()

        self.dragDropTask = None

        self.firstFrame = self.createcomponent(
            'firstFrame', (), None,
            DirectFrame, (self,),
            frameSize = (-0.45, 0.45, -1, 1),
            pos = (-0.55, 0, 0),)

        self.secondFrame = self.createcomponent(
            'secondFrame', (), None,
            DirectFrame, (self,),
            frameSize = (-0.45, 0.45, -1, 1),
            pos = (0.55, 0, 0),)

        self.splitter = self.createcomponent(
            'splitter', (), None,
            DirectFrame, (self,),
            numStates = 2,
            frameSize = (-0.01, 0.01, -1, 1),
            frameColor = self['splitterColor'],
            text = "<\n>",
            text_scale = 0.05,
            text_align = TextNode.ACenter,
            state = 'normal')
        self.splitter.bind(DGG.ENTER, self.enter)
        self.splitter.bind(DGG.EXIT, self.exit)
        self.splitter.bind(DGG.B1PRESS, self.dragStart)
        self.splitter.bind(DGG.B1RELEASE, self.dragStop)

        self.setSplitter(self['splitterWidth'])

        self.refresh()

    def setSplitter(self, width=0.02):
        if not hasattr(self, "splitter"): return
        if self['showSplitter'] == False:
            self.splitter.hide()
            self['splitterWidth'] = 0
        else:
            self.splitter.show()
            self['splitterWidth'] = width


    def refresh(self):
        """
        Recalculate the position of every item in this panel and set the frame-
        size of the panel accordingly if auto update is enabled.
        """
        # sanity check so we don't get here to early
        if not hasattr(self, "bounds"): return

        width = DGH.getRealWidth(self)
        height = DGH.getRealHeight(self)

        if self["orientation"] == DGG.HORIZONTAL:

            #TODO: BETTER CALCULATION NEEDED.
            # This should respect the frame size in every direction. pixel2d has 0 at left with 0 to +x but could also differ while aspect usually has 0 at the center ranging -x to +x
            # So we have to take the left and right edge of the actual frame into account. This may also need to be respected at the moving part, though this might already work since it's
            # respective to the frame as it is its parent.
            # maybe we can also implement multi splits
            if self["pixel2d"]:
                splitterPosInPercent = 1 - self["splitterPos"]/width
                leftWidth = width * (1-splitterPosInPercent) - (self["splitterWidth"] / 2)
                rightWidth = width * splitterPosInPercent - (self["splitterWidth"] / 2)
            else:
                leftWidth = (width / 2) + self["splitterPos"] - (self["splitterWidth"] / 2)
                rightWidth = (width / 2) - self["splitterPos"] - (self["splitterWidth"] / 2)

            self.firstFrame["frameSize"] = (-leftWidth/2, leftWidth/2, self["frameSize"][2], self["frameSize"][3])
            self.firstFrame.setX((-leftWidth / 2) - (self["splitterWidth"] / 2) + self["splitterPos"])
            self.firstFrame.setZ(0)
            self.secondFrame["frameSize"] = (-rightWidth/2, rightWidth/2, self["frameSize"][2], self["frameSize"][3])
            self.secondFrame.setX((rightWidth / 2) + (self["splitterWidth"] / 2) + self["splitterPos"])
            self.secondFrame.setZ(0)

            self.splitter["frameSize"] = (-self["splitterWidth"]/2, self["splitterWidth"]/2, self["frameSize"][2], self["frameSize"][3])
            self.splitter.setX(self["splitterPos"])
            self.splitter["text_roll"] = 0

        elif self["orientation"] == DGG.VERTICAL:
            if self["pixel2d"]:
                splitterPosInPercent = 1 - self["splitterPos"]/height
                topHeight = height * splitterPosInPercent - (self["splitterWidth"] / 2)
                bottomHeight = height * (1-splitterPosInPercent) - (self["splitterWidth"] / 2)
            else:
                topHeight = (height / 2) - self["splitterPos"] - (self["splitterWidth"] / 2)
                bottomHeight = (height / 2) + self["splitterPos"] - (self["splitterWidth"] / 2)

            self.firstFrame["frameSize"] = (self["frameSize"][0], self["frameSize"][1], -topHeight/2, topHeight/2)
            self.firstFrame.setX(0)
            self.firstFrame.setZ((topHeight / 2) + (self["splitterWidth"] / 2) + self["splitterPos"])
            self.secondFrame["frameSize"] = (self["frameSize"][0], self["frameSize"][1], -bottomHeight/2, bottomHeight/2)
            self.secondFrame.setX(0)
            self.secondFrame.setZ((-bottomHeight / 2) - (self["splitterWidth"] / 2) + self["splitterPos"])

            self.splitter["frameSize"] = (self["frameSize"][0], self["frameSize"][1], -self["splitterWidth"]/2, self["splitterWidth"]/2)
            self.splitter.setZ(self["splitterPos"])
            self.splitter["text_roll"] = 90


        base.messenger.send(self.uniqueName("update-size"))
        if self['firstFrameUpdateSizeFunc'] is not None:
            self['firstFrameUpdateSizeFunc']()
        if self['secondFrameUpdateSizeFunc'] is not None:
            self['secondFrameUpdateSizeFunc']()

    def enter(self, event):
        self.splitter["frameColor"] = self['splitterHighlightColor']

    def exit(self, event):
        self.splitter["frameColor"] = self['splitterColor']

    def dragStart(self, event):
        """
        Start dragging the window around
        """
        if self.dragDropTask is not None:
            # remove any existing tasks
            taskMgr.remove(self.dragDropTask)

        self.splitter["frameColor"] = self['splitterHighlightColor']

        # get the windows position as seen from render2d
        vWidget2render2d = self.splitter.getPos(render2d)
        # get the mouse position as seen from render2d
        vMouse2render2d = Point3(event.getMouse()[0], 0, event.getMouse()[1])
        # calculate the vector between the mosue and the window
        editVec = Vec3(vWidget2render2d - vMouse2render2d)
        # create the task and store the values in it, so we can use it in there
        self.dragDropTask = taskMgr.add(self.dragTask, self.taskName("dragDropTask"))
        self.dragDropTask.editVec = editVec
        self.dragDropTask.mouseVec = vMouse2render2d

    def dragTask(self, t):
        """
        Calculate the new window position ever frame
        """
        # chec if we have the mouse
        mwn = base.mouseWatcherNode
        if mwn.hasMouse():
            # get the mouse position
            vMouse2render2d = Point3(mwn.getMouse()[0], 0, mwn.getMouse()[1])
            # calculate the new position using the mouse position and the start
            # vector of the window
            newPos = vMouse2render2d + t.editVec
            # Now set the new windows new position
            if self["orientation"] == DGG.HORIZONTAL:
                newPos.setY(0)
                newPos.setZ(self.splitter.getZ(render2d))
                self.splitter.setPos(render2d, newPos)
                minLeft = self['firstFrameMinSize'] if self['firstFrameMinSize'] is not None else 0
                minRight = self['secondFrameMinSize'] if self['secondFrameMinSize'] is not None else 0
                if self.splitter.getX() - self["splitterWidth"] / 2 < self["frameSize"][0] + minLeft:
                    self.splitter.setX(self["frameSize"][0] + self["splitterWidth"] / 2 + minLeft)
                elif self.splitter.getX() + self["splitterWidth"] / 2 > self["frameSize"][1] - minRight:
                    self.splitter.setX(self["frameSize"][1] - self["splitterWidth"] / 2 - minRight)
                self["splitterPos"] = self.splitter.getX()

            elif self["orientation"] == DGG.VERTICAL:
                newPos.setX(self.splitter.getX(render2d))
                newPos.setY(0)
                self.splitter.setPos(render2d, newPos)
                minTop = self['firstFrameMinSize'] if self['firstFrameMinSize'] is not None else 0
                minBottom = self['secondFrameMinSize'] if self['secondFrameMinSize'] is not None else 0
                if self.splitter.getZ() - self["splitterWidth"] / 2 < self["frameSize"][2] + minTop:
                    self.splitter.setZ(self["frameSize"][2] + self["splitterWidth"] / 2 + minTop)
                elif self.splitter.getZ() + self["splitterWidth"] / 2 > self["frameSize"][3] - minBottom:
                    self.splitter.setZ(self["frameSize"][3] - self["splitterWidth"] / 2 - minBottom)
                self["splitterPos"] = self.splitter.getZ()
            self.refresh()
            self.splitter["frameColor"] = self['splitterHighlightColor']
        return t.cont

    def dragStop(self, event):
        """
        Stop dragging the splitter around
        """
        # kill the drag and drop task
        taskMgr.remove(self.dragDropTask)
        self.splitter["frameColor"] = self['splitterColor']
        self.refresh()



"""Implements a pop-up menu containing multiple clickable options."""

__all__ = ['DirectOptionMenu']

from panda3d.core import *
from direct.gui import DirectGuiGlobals as DGG
#import DirectGuiGlobalsExtra as DGG #<- TODO: Why doesn't this work as expected
DGG.MWUP = PGButton.getPressPrefix() + MouseButton.wheel_up().getName() + '-'
DGG.MWDOWN = PGButton.getPressPrefix() + MouseButton.wheel_down().getName() + '-'
DGG.LEFT = "left"
DGG.RIGHT = "right"
DGG.ABOVE = "above"
DGG.BELOW = "below"
from direct.gui.DirectButton import *
from direct.gui.DirectLabel import *
from direct.gui.DirectFrame import *

class DirectOptionMenu(DirectButton):
    """
    DirectOptionMenu(parent) - Create a DirectButton which pops up a
    menu which can be used to select from a list of items.
    Execute button command (passing the selected item through) if defined
    To cancel the popup menu click anywhere on the screen outside of the
    popup menu.  No command is executed in this case.
    """
    def __init__(self, parent = None, **kw):
        # Inherits from DirectButton
        optiondefs = (
            # List of items to display on the popup menu
            ('items',       [],             self.setItems),
            # Initial item to display on menu button
            # Can be an integer index or the same string as the button
            ('initialitem', None,           DGG.INITOPT),
            # Amount of padding to place around popup button indicator
            ('popupMarkerBorder', (.1, .1), None),
            # The initial position of the popup marker
            ('popupMarker_pos', None, None),
            # The position of the popup menu
            # possible positions: left, above, right, below
            ('popupMenuLocation', None, None),
            # Background color to use to highlight popup menu items
            ('highlightColor', (.5, .5, .5, 1), None),
            # Extra scale to use on highlight popup menu items
            ('highlightScale', (1, 1), None),
            # Alignment to use for text on popup menu button
            # Changing this breaks button layout
            ('text_align',  TextNode.ALeft, None),
            # Remove press effect because it looks a bit funny
            ('pressEffect',     0,          DGG.INITOPT),
           )
        # Merge keyword options with default options
        self.defineoptions(kw, optiondefs)
        # Initialize superclasses
        DirectButton.__init__(self, parent)
        # Record any user specified frame size
        self.initFrameSize = self['frameSize']
        # Create a small rectangular marker to distinguish this button
        # as a popup menu button
        self.popupMarker = self.createcomponent(
            'popupMarker', (), None,
            DirectFrame, (self,),
            frameSize = (-0.5, 0.5, -0.2, 0.2),
            scale = 0.4,
            relief = DGG.RAISED)
        # Record any user specified popup marker position
        self.initPopupMarkerPos = self['popupMarker_pos']
        # This needs to popup the menu too
        self.popupMarker.bind(DGG.B1PRESS, self.showPopupMenu)
        # Check if item is highlighted on release and select it if it is
        self.popupMarker.bind(DGG.B1RELEASE, self.selectHighlightedIndex)
        # Make popup marker have the same click sound
        if self['clickSound']:
            self.popupMarker.guiItem.setSound(
                DGG.B1PRESS + self.popupMarker.guiId, self['clickSound'])
        else:
            self.popupMarker.guiItem.clearSound(DGG.B1PRESS + self.popupMarker.guiId)
        # This is created when you set the menu's items
        self.popupMenu = None
        self.selectedIndex = None
        self.highlightedIndex = None
        if 'item_text_scale' in kw:
            self.prevItemTextScale = kw['item_text_scale']
        else:
            self.prevItemTextScale = (1,1)
        # A big screen encompassing frame to catch the cancel clicks
        self.cancelFrame = self.createcomponent(
            'cancelframe', (), None,
            DirectFrame, (self,),
            frameSize = (-1, 1, -1, 1),
            frameColor = (1, 0, 0, 1),
            relief = None,
            state = 'normal')
        # Make sure this is on top of all the other widgets
        self.cancelFrame.setBin('gui-popup', 0)
        self.cancelFrame.bind(DGG.B1PRESS, self.hidePopupMenu)
        # Default action on press is to show popup menu
        self.bind(DGG.B1PRESS, self.showPopupMenu)
        # Check if item is highlighted on release and select it if it is
        self.bind(DGG.B1RELEASE, self.selectHighlightedIndex)
        # Call option initialization functions
        self.initialiseoptions(DirectOptionMenu)
        # Need to call this since we explicitly set frame size
        self.resetFrameSize()

    def setItems(self):
        """
        self['items'] = itemList
        Create new popup menu to reflect specified set of items
        """
        # Remove old component if it exits
        if self.popupMenu != None:
            self.destroycomponent('popupMenu')
        # Create new component
        self.popupMenu = self.createcomponent('popupMenu', (), None,
                                              DirectFrame,
                                              (self,),
                                              relief = 'raised',
                                              )
        # Make sure it is on top of all the other gui widgets
        self.popupMenu.setBin('gui-popup', 0)
        if not self['items']:
            return
        # Create a new component for each item
        # Find the maximum extents of all items
        itemIndex = 0
        self.minX = self.maxX = self.minZ = self.maxZ = None
        for item in self['items']:
            c = self.createcomponent(
                'item%d' % itemIndex, (), 'item',
                DirectButton, (self.popupMenu,),
                text = item, text_align = TextNode.ALeft,
                command = lambda i = itemIndex: self.set(i))
            bounds = c.getBounds()
            if self.minX == None:
                self.minX = bounds[0]
            elif bounds[0] < self.minX:
                self.minX = bounds[0]
            if self.maxX == None:
                self.maxX = bounds[1]
            elif bounds[1] > self.maxX:
                self.maxX = bounds[1]
            if self.minZ == None:
                self.minZ = bounds[2]
            elif bounds[2] < self.minZ:
                self.minZ = bounds[2]
            if self.maxZ == None:
                self.maxZ = bounds[3]
            elif bounds[3] > self.maxZ:
                self.maxZ = bounds[3]
            itemIndex += 1
        # Calc max width and height
        self.maxWidth = self.maxX - self.minX
        self.maxHeight = self.maxZ - self.minZ
        # Adjust frame size for each item and bind actions to mouse events
        for i in range(itemIndex):
            item = self.component('item%d' %i)
            # So entire extent of item's slot on popup is reactive to mouse
            item['frameSize'] = (self.minX, self.maxX, self.minZ, self.maxZ)
            # Move it to its correct position on the popup
            item.setPos(-self.minX, 0, -self.maxZ - i * self.maxHeight)
            item.bind(DGG.B1RELEASE, self.hidePopupMenu)
            # Highlight background when mouse is in item
            item.bind(DGG.WITHIN,
                      lambda x, i=i, item=item:self._highlightItem(item, i))
            # Restore specified color upon exiting
            fc = item['frameColor']
            item.bind(DGG.WITHOUT,
                      lambda x, item=item, fc=fc: self._unhighlightItem(item, fc))
            item.bind(DGG.MWDOWN, self.scrollPopUpMenu, [-1])
            item.bind(DGG.MWUP, self.scrollPopUpMenu, [1])
        # Set popup menu frame size to encompass all items
        f = self.component('popupMenu')
        f['frameSize'] = (0, self.maxWidth, -self.maxHeight * itemIndex, 0)
        # HACK: Set the user defined popup menu relief here so we don't
        # break the bounds calculation.
        self.popupMenu.setRelief(self['popupMenu_relief'])

        # Determine what initial item to display and set text accordingly
        if self['initialitem']:
            self.set(self['initialitem'], fCommand = 0)
        else:
            # No initial item specified, just use first item
            self.set(0, fCommand = 0)

        # Position popup Marker to the right of the button
        pm = self.popupMarker
        pmw = (pm.getWidth() * pm.getScale()[0] +
               2 * self['popupMarkerBorder'][0])
        if self.initFrameSize:
            # Use specified frame size
            bounds = list(self.initFrameSize)
        else:
            # Or base it upon largest item
            bounds = [self.minX, self.maxX, self.minZ, self.maxZ]
        if self.initPopupMarkerPos:
            # Use specified position
            pmPos = list(self.initPopupMarkerPos)
        else:
            # Or base the position on the frame size.
            pmPos = [bounds[1] + pmw/2.0, 0, bounds[2] + (bounds[3] - bounds[2])/2.0]
        pm.setPos(pmPos[0], pmPos[1], pmPos[2])
        # Adjust popup menu button to fit all items (or use user specified
        # frame size
        bounds[1] += pmw
        self['frameSize'] = (bounds[0], bounds[1], bounds[2], bounds[3])
        # Set initial state
        self.hidePopupMenu()
        self.bind(DGG.MWDOWN, self.scrollPopUpMenu, [-1])
        self.bind(DGG.MWUP, self.scrollPopUpMenu, [1])
        self.cancelFrame.bind(DGG.MWDOWN, self.scrollPopUpMenu, [-1])
        self.cancelFrame.bind(DGG.MWUP, self.scrollPopUpMenu, [1])

    def showPopupMenu(self, event = None):
        """
        Make popup visible and try to position it just to right of
        mouse click with currently selected item aligned with button.
        Adjust popup position if default position puts it outside of
        visible screen region
        """

        # Needed attributes (such as minZ) won't be set unless the user has specified
        # items to display. Let's assert that we've given items to work with.
        items = self['items']
        assert items and len(items) > 0, 'Cannot show an empty popup menu! You must add items!'

        # Show the menu
        self.popupMenu.show()
        # Make sure its at the right scale
        self.popupMenu.setScale(self, VBase3(1))
        # Compute bounds
        b = self.getBounds()
        fb = self.popupMenu.getBounds()

        if self['popupMenuLocation'] == DGG.RIGHT or self['popupMenuLocation'] == None:
            # This is the default to not break existing applications
            # Position menu at midpoint of button
            xPos = (b[1] - b[0])/2.0 - fb[0]
        elif self['popupMenuLocation'] == DGG.LEFT:
            # Position to the left
            xPos = -fb[1] + (b[1] - b[0])/2.0
        else:
            # position to line up with the left edge if the menu is above or below
            xPos = b[0]
        self.popupMenu.setX(self, xPos)

        if self['popupMenuLocation'] == DGG.ABOVE:
            # Try to set height to line up selected item with button
            self.popupMenu.setZ(
                self, self.maxZ - fb[2])
        elif self['popupMenuLocation'] == DGG.BELOW:
            # Try to set height to line up selected item with button
            self.popupMenu.setZ(
                self, self.minZ)
        else:
            # Try to set height to line up selected item with button
            self.popupMenu.setZ(
                self, self.minZ + (self.selectedIndex + 1)*self.maxHeight)
        # Make sure the whole popup menu is visible
        pos = self.popupMenu.getPos(render2d)
        scale = self.popupMenu.getScale(render2d)
        # How are we doing relative to the right side of the screen
        maxX = pos[0] + fb[1] * scale[0]
        if maxX > 1.0:
            # Need to move menu to the left
            self.popupMenu.setX(render2d, pos[0] + (1.0 - maxX))
        # How are we doing relative to the right side of the screen
        minX = pos[0]
        if minX < -1.0:
            # Need to move menu to the right
            self.popupMenu.setX(render2d, -1 )
        # How about up and down?
        minZ = pos[2] + fb[2] * scale[2]
        maxZ = pos[2] + fb[3] * scale[2]
        if minZ < -1.0:
            # Menu too low, move it up
            self.popupMenu.setZ(render2d, pos[2] + (-1.0 - minZ))

            # recheck the top position once repositioned
            pos = self.popupMenu.getPos(render2d)
            maxZ = pos[2] + fb[3] * scale[2]
            if maxZ > 1.0:
                # Menu too large to show on screen entirely
                # Try to set height to line up selected item with button
                self.popupMenu.setZ(
                    self, self.minZ + (self.selectedIndex + 1)*self.maxHeight)
        elif maxZ > 1.0:
            # Menu too high, move it down
            self.popupMenu.setZ(render2d, pos[2] + (1.0 - maxZ))
            # recheck the top position once repositioned
            pos = self.popupMenu.getPos(render2d)
            minZ = pos[2] + fb[2] * scale[2]
            if minZ < -1.0:
                # Menu too large to show on screen entirely
                # Try to set height to line up selected item with button
                self.popupMenu.setZ(
                    self, self.minZ + (self.selectedIndex + 1)*self.maxHeight)
        # Also display cancel frame to catch clicks outside of the popup
        self.cancelFrame.show()
        # Position and scale cancel frame to fill entire window
        self.cancelFrame.setPos(render2d, 0, 0, 0)
        self.cancelFrame.setScale(render2d, 1, 1, 1)

    def hidePopupMenu(self, event = None):
        """ Put away popup and cancel frame """
        self.popupMenu.hide()
        self.cancelFrame.hide()

    def scrollPopUpMenu(self, direction, event = None):
        """ Scroll the item frame up and down depending on the direction
        which must be a nummeric value. A positive value will scroll up
        while a negative value will scroll down. It will only work if
        items are out of bounds of the window """
        fb = self.popupMenu.getBounds()
        pos = self.popupMenu.getPos(render2d)
        scale = self.popupMenu.getScale(render2d)

        minZ = pos[2] + fb[2] * scale[2]
        maxZ = pos[2] + fb[3] * scale[2]
        if (minZ < -1.0 and direction > 0) or (maxZ > 1.0 and direction < 0):
            oldZ = self.popupMenu.getZ()
            self.popupMenu.setZ(oldZ + direction * self.maxHeight)

    def _highlightItem(self, item, index):
        """ Set frame color of highlighted item, record index """
        self.prevItemTextScale = item['text_scale']
        item['frameColor'] = self['highlightColor']
        #item['frameSize'] = (self['highlightScale'][0]*self.minX, self['highlightScale'][0]*self.maxX, self['highlightScale'][1]*self.minZ, self['highlightScale'][1]*self.maxZ)
        item['text_scale'] = self['highlightScale']
        self.highlightedIndex = index

    def _unhighlightItem(self, item, frameColor):
        """ Clear frame color, clear highlightedIndex """
        item['frameColor'] = frameColor
        item['frameSize'] = (self.minX, self.maxX, self.minZ, self.maxZ)
        item['text_scale'] = self.prevItemTextScale
        self.highlightedIndex = None

    def selectHighlightedIndex(self, event = None):
        """
        Check to see if item is highlighted (by cursor being within
        that item).  If so, selected it.  If not, do nothing
        """
        if self.highlightedIndex is not None:
            self.set(self.highlightedIndex)
            self.hidePopupMenu()

    def index(self, index):
        intIndex = None
        if isinstance(index, int):
            intIndex = index
        elif index in self['items']:
            i = 0
            for item in self['items']:
                if item == index:
                    intIndex = i
                    break
                i += 1
        return intIndex

    def set(self, index, fCommand = 1):
        # Item was selected, record item and call command if any
        newIndex = self.index(index)
        if newIndex is not None:
            self.selectedIndex = newIndex
            item = self['items'][self.selectedIndex]
            self['text'] = item
            if fCommand and self['command']:
                # Pass any extra args to command
                self['command'](*[item] + self['extraArgs'])

    def get(self):
        """ Get currently selected item """
        return self['items'][self.selectedIndex]

    def commandFunc(self, event):
        """
        Override popup menu button's command func
        Command is executed in response to selecting menu items
        """
        pass



"""This module contains the DirectAutoSizer class."""

__all__ = ['DirectAutoSizer']

import inspect
from panda3d.core import *
from direct.gui import DirectGuiGlobals as DGG
from direct.gui.DirectFrame import DirectFrame
from direct.gui.DirectGuiBase import DirectGuiWidget
from . import DirectGuiHelper as DGH
from direct.showbase import ShowBaseGlobal

class DirectAutoSizer(DirectFrame):
    """
    A frame to Automatically resize the given other DirectGui element
    """
    def __init__(self, parent = None, child = None, **kw):
        optiondefs = (
            ('extendHorizontal', True,      None),
            ('extendVertical',   True,      None),
            ('minSize',          (0, 0, 0, 0), self.refresh),
            ('maxSize',          (0, 0, 0, 0), self.refresh),
            ('updateOnWindowResize', True,  self.setUpdateOnWindowResize),
            ('childUpdateSizeFunc', None,   None),
            ('parentGetSizeFunction', None, None),
            ('parentGetSizeExtraArgs', None, None),

            ('suppressMouse',  0,           None),
            )
        # Merge keyword options with default options
        self.defineoptions(kw, optiondefs)

        # Initialize superclasses
        DirectFrame.__init__(self, parent)

        self.parentObject = parent
        self.child = child
        if child is not None:
            child.reparentTo(self)

        # Call option initialization functions
        self.initialiseoptions(DirectAutoSizer)

    def setChild(self, child):
        if self.child is not None:
            self.child.detachNode()
        self.child = child
        self.child.reparentTo(self)
        self.refresh()

    def removeChild(self):
        if self.child is not None:
            self.child.detachNode()
        self.child = None

    def setUpdateOnWindowResize(self):
        if self['updateOnWindowResize']:
            # Make sure the sizer scales with window size changes
            self.screenSize = base.getSize()
            self.accept('window-event', self.windowEventHandler)
        else:
            # stop updating on window resize events
            self.ignore('window-event')
            self.screenSize = None

    def windowEventHandler(self, window=None):
        if window != base.win:
            # This event isn't about our window.
            return

        #if window is not None: # window is none if panda3d is not started
        #    if self.screenSize == base.getSize():
        #        return
        #self.screenSize = base.getSize()
        taskMgr.doMethodLater(0.25, self.refresh, "delayed sizer refresh", extraArgs=[])
        #self.refresh()

    def refresh(self):
        """Resize the sizer and its child element"""
        if self.child is None:
            return

        # store left/right/bottom/top
        l=r=b=t=0

        # dependent on our parent, make sure the size values are set correct
        if self["parentGetSizeFunction"] is not None:
            # we have a user defined way to get the size, this should be treated
            # most important
            size = self["parentGetSizeFunction"](*self["parentGetSizeExtraArgs"])

            l=size[0]
            r=size[1]
            b=size[2]
            t=size[3]
        elif self.parentObject is None or self.parentObject == ShowBaseGlobal.aspect2d:
            # the default parent of directGui widgets
            l = base.a2dLeft
            r = base.a2dRight
            b = base.a2dBottom
            t = base.a2dTop
        elif self.parentObject == base.pixel2d:
            # we are parented to pixel2d
            xsize, ysize = base.getSize()
            l = 0
            r = xsize
            b = -ysize
            t = 0
        elif DirectGuiWidget in inspect.getmro(type(self.parentObject)):
            # We have a "normal" DirectGui widget here
            bounds = self.parentObject.bounds
            l, r, b, t = bounds
        else:
            # We are parented to something else, probably a nodepath
            self.parentObject.node().setBoundsType(BoundingVolume.BT_box)

            ll = LPoint3()
            ur = LPoint3()

            self.parentObject.calcTightBounds(ll, ur, render)

            l=ll.getX()
            r=ur.getX()
            b=ll.getZ()
            t=ur.getZ()

        childSize = self.child['frameSize']
        if childSize is None:
            childSize = self.child.getBounds()

        if type(self.child["scale"]) == LVecBase3f:
            childScale = self.child["scale"]
        elif self.child["scale"] is not None:
            try:
                childScale = LVecBase3f(self.child["scale"])
            except:
                childScale = self.child["scale"]
        else:
            childScale = LVecBase3f(1.0)

        if not self['extendHorizontal']:
            l = childSize[0] * childScale.getX()
            r = childSize[1] * childScale.getX()

        if not self['extendVertical']:
            b = childSize[2] * childScale.getY()
            t = childSize[3] * childScale.getY()

        #TODO: Better check for positive/negative numbers
        if l > self['minSize'][0]: l = self['minSize'][0]
        if r < self['minSize'][1]: r = self['minSize'][1]
        if b > self['minSize'][2]: b = self['minSize'][2]
        if t < self['minSize'][3]: t = self['minSize'][3]

        if self['maxSize'] is not None and self['maxSize'] != (0, 0, 0, 0):
            if l < self['maxSize'][0]: l = self['maxSize'][0]
            if r > self['maxSize'][1]: r = self['maxSize'][1]
            if b < self['maxSize'][2]: b = self['maxSize'][2]
            if t > self['maxSize'][3]: t = self['maxSize'][3]

        # recalculate size according to position
        l -= self.child.getX()
        r += self.child.getX()
        b -= self.child.getZ()
        t += self.child.getZ()

        # actual resizing of our child element
        self.child["frameSize"] = [l/childScale.getX(),r/childScale.getX(),b/childScale.getY(),t/childScale.getY()]
        self["frameSize"] = self.child["frameSize"]

        base.messenger.send(self.uniqueName("update-size"))
        if self['childUpdateSizeFunc'] is not None:
            self['childUpdateSizeFunc']()



"""This module contains the DirectDiagram class."""

__all__ = ['DirectDiagram']

import math
from panda3d.core import *
from direct.gui import DirectGuiGlobals as DGG
from direct.gui.DirectFrame import DirectFrame
from direct.gui.DirectLabel import DirectLabel
from . import DirectGuiHelper as DGH
from direct.directtools.DirectGeometry import LineNodePath

class DirectDiagram(DirectFrame):

    def __init__(self, parent = None, **kw):
        optiondefs = (
            # Define type of DirectGuiWidget
            ('data',           [],          self.refresh),
            ('numPosSteps',     0,          self.refresh),
            ('numPosStepsStep', 1,          self.refresh),
            ('numNegSteps',     0,          self.refresh),
            ('numNegStepsStep', 1,          self.refresh),
            ('numtextScale',    0.05,       self.refresh),
            ('showDataNumbers', False,      self.refresh),
            ('dataNumtextScale',0.05,       self.refresh),
            ('stepAccuracy',    2,          self.refresh),
            ('stepFormat',      float,      self.refresh),
            ('numberAreaWidth', 0.15,          self.refresh),
            #('numStates',      1,           None),
            #('state',          DGG.NORMAL,  None),
            )
        # Merge keyword options with default options
        self.defineoptions(kw, optiondefs)

        self.lines = None
        self.measureLines = None
        self.centerLine = None
        self.xDescriptions = []
        self.points = []

        # Initialize superclasses
        DirectFrame.__init__(self, parent)

        # Call option initialization functions
        self.initialiseoptions(DirectDiagram)


    def refresh(self):
        # sanity check so we don't get here to early
        if not hasattr(self, "bounds"): return
        self.frameInitialiseFunc()

        textLeftSizeArea = self['numberAreaWidth']
        # get the left and right edge of our frame
        left = DGH.getRealLeft(self)
        right = DGH.getRealRight(self)
        diagramLeft = left + textLeftSizeArea

        xStep = (DGH.getRealWidth(self) - textLeftSizeArea) / (len(self['data'])-1)
        posYRes = DGH.getRealTop(self) / (self['numPosSteps'] if self['numPosSteps'] > 0 else max(self['data']))
        negYRes = DGH.getRealBottom(self) / (-self['numNegSteps'] if self['numNegSteps'] > 0 else min(self['data']))

        # remove old content
        if self.lines is not None:
            self.lines.removeNode()

        if self.measureLines is not None:
            self.measureLines.removeNode()

        if self.centerLine is not None:
            self.centerLine.removeNode()

        for text in self.xDescriptions:
            text.removeNode()
        self.xDescriptions = []

        for text in self.points:
            text.removeNode()
        self.points = []

        # prepare the line drawings
        self.lines = LineNodePath(parent=self, thickness=3.0, colorVec=(1, 0, 0, 1))
        self.measureLines = LineNodePath(parent=self, thickness=1.0, colorVec=(0, 0, 0, 1))
        self.centerLine = LineNodePath(parent=self, thickness=2.0, colorVec=(0, 0, 0, 1))

        # draw the center line
        self.centerLine.reset()
        self.centerLine.drawLines([((diagramLeft, 0, 0), (right, 0, 0))])
        self.centerLine.create()

        self.xDescriptions.append(
            self.createcomponent(
                'value0', (), None,
                DirectLabel, (self,),
                text = "0",
                text_scale = self['numtextScale'],
                text_align = TextNode.ARight,
                pos = (diagramLeft, 0, -0.01),
                relief = None,
                state = 'normal'))

        # calculate the positive measure lines and add the numbers
        measureLineData = []
        numSteps = (self['numPosSteps'] if self['numPosSteps'] >0 else math.floor(max(self['data']))) + 1
        for i in range(1, numSteps, self['numPosStepsStep']):
            measureLineData.append(
                (
                    (diagramLeft, 0, i*posYRes),
                    (right, 0, i*posYRes)
                )
            )

            calcBase = 1 / DGH.getRealTop(self)
            maxData = self['numPosSteps'] if self['numPosSteps'] >0 else max(self['data'])
            value = self['stepFormat'](round(i * posYRes * calcBase * maxData, self['stepAccuracy']))
            y = i*posYRes
            self.xDescriptions.append(
                self.createcomponent(
                    'value{}'.format(value), (), None,
                    DirectLabel, (self,),
                    text = str(value),
                    text_scale = self['numtextScale'],
                    text_align = TextNode.ARight,
                    pos = (diagramLeft, 0, y-0.025),
                    relief = None,
                    state = 'normal'))

        # calculate the negative measure lines and add the numbers
        numSteps = (self['numNegSteps'] if self['numNegSteps'] >0 else math.floor(abs(min(self['data'])))) + 1
        for i in range(1, numSteps, self['numNegStepsStep']):
            measureLineData.append(
                (
                    (diagramLeft, 0, -i*negYRes),
                    (right, 0, -i*negYRes)
                )
            )

            calcBase = 1 / DGH.getRealBottom(self)
            maxData = self['numPosSteps'] if self['numPosSteps'] >0 else max(self['data'])
            value = self['stepFormat'](round(i * negYRes * calcBase * maxData, self['stepAccuracy']))
            y = -i*negYRes
            self.xDescriptions.append(
                self.createcomponent(
                    'value{}'.format(value), (), None,
                    DirectLabel, (self,),
                    text = str(value),
                    text_scale = self['numtextScale'],
                    text_align = TextNode.ARight,
                    pos = (diagramLeft, 0, y+0.01),
                    relief = None,
                    state = 'normal'))

        # Draw the lines
        self.measureLines.reset()
        self.measureLines.drawLines(measureLineData)
        self.measureLines.create()

        lineData = []
        for i in range(1, len(self['data'])):
            yResA = posYRes if self['data'][i-1] >= 0 else negYRes
            yResB = posYRes if self['data'][i] >= 0 else negYRes
            lineData.append(
                (
                    # Point A
                    (diagramLeft+(i-1)*xStep, 0, self['data'][i-1] * yResA),
                    # Point B
                    (diagramLeft+i*xStep, 0, self['data'][i] * yResB)
                )
            )

            if (self['showDataNumbers']):
                value = round(self['data'][i-1], self['stepAccuracy'])
                self.points.append(
                    self.createcomponent(
                        'value{}'.format(value), (), None,
                        DirectLabel, (self,),
                        text = str(value),
                        text_scale = self['dataNumtextScale'],
                        text_align = TextNode.ARight,
                        pos = (diagramLeft+(i-1)*xStep, 0, self['data'][i-1] * yResA),
                        relief = None,
                        state = 'normal'))

        # Draw the lines
        self.lines.reset()
        self.lines.drawLines(lineData)
        self.lines.create()



"""This module contains the DirectScrolledWindowFrame class.

It will create a scrolled frame with an extra frame on top of it
"""

__all__ = ['DirectScrolledWindowFrame']

from panda3d.core import *
from direct.gui import DirectGuiGlobals as DGG
from direct.gui.DirectFrame import DirectFrame
from direct.gui.DirectButton import DirectButton
from direct.gui.DirectScrolledFrame import DirectScrolledFrame

class DirectScrolledWindowFrame(DirectScrolledFrame):
    """
    A moveable window with a scrolled content frame
    """
    def __init__(self, parent = None, **kw):
        optiondefs = (
            # Define type of DirectGuiWidget
            # The height of the area to drag the widget around
            ('dragAreaHeight',              0.1, None),
            ('resortOnDrag',               True, None),
            ('showClose',                  True, None),
            ('closeButtonPosition',     'Right', None),
            ('closeButtonScale',           0.05, None)
            )
        # Merge keyword options with default options
        self.defineoptions(kw, optiondefs)

        # Initialize superclasses
        DirectScrolledFrame.__init__(self, parent)

        # Call option initialization functions
        self.initialiseoptions(DirectScrolledWindowFrame)

        self.dragDropTask = None

        b = self.bounds
        c = self.createcomponent(
            'dragFrame', (), 'dragFrame',
            DirectFrame,
            # set the parent of the frame to this class
            (self,),
            state=DGG.NORMAL,
            suppressMouse=True,
            frameColor=(0.5,0.5,0.5,1),
            relief=1,
            pos=(0,0,b[3]),
            # set the size
            frameSize=(b[0],b[1],0, self['dragAreaHeight']))

        c.bind(DGG.B1PRESS, self.dragStart)
        c.bind(DGG.B1RELEASE, self.dragStop)

        scale = self['closeButtonScale']
        pos = (0,0,self['dragAreaHeight']*0.5)
        if self['closeButtonPosition'] == 'Right':
            pos = (b[1]-scale*0.5,0,self['dragAreaHeight']*0.5)
        elif self['closeButtonPosition'] == 'Left':
            pos = (b[0]+scale*0.5,0,self['dragAreaHeight']*0.5)
        closeBtn = self.createcomponent(
            'closeButton', (), 'closeButton',
            DirectButton,
            (c,),
            text='x',
            scale=scale,
            pos=pos,
            command=self.destroy)

    def dragStart(self, event):
        """
        Start dragging the window around
        """
        if self.dragDropTask is not None:
            # remove any existing tasks
            taskMgr.remove(self.dragDropTask)

        if self['resortOnDrag']:
            self.reparentTo(self.parent, 0)

        # get the windows position as seen from render2d
        vWidget2render2d = self.getPos(render2d)
        # get the mouse position as seen from render2d
        vMouse2render2d = Point3(event.getMouse()[0], 0, event.getMouse()[1])
        # calculate the vector between the mosue and the window
        editVec = Vec3(vWidget2render2d - vMouse2render2d)
        # create the task and store the values in it, so we can use it in there
        self.dragDropTask = taskMgr.add(self.dragTask, self.taskName("dragDropTask"))
        self.dragDropTask.editVec = editVec
        self.dragDropTask.mouseVec = vMouse2render2d

    def dragTask(self, t):
        """
        Calculate the new window position ever frame
        """
        # chec if we have the mouse
        mwn = base.mouseWatcherNode
        if mwn.hasMouse():
            # get the mouse position
            vMouse2render2d = Point3(mwn.getMouse()[0], 0, mwn.getMouse()[1])
            # calculate the new position using the mouse position and the start
            # vector of the window
            newPos = vMouse2render2d + t.editVec
            # Now set the new windows new position
            self.setPos(render2d, newPos)
        return t.cont

    def dragStop(self, event):
        """
        Stop dragging the window around
        """
        # kill the drag and drop task
        taskMgr.remove(self.dragDropTask)



"""This module contains the DirectBoxSizer class."""

__all__ = ['DirectGridSizer']

from panda3d.core import *
from direct.gui import DirectGuiGlobals as DGG
from direct.gui.DirectFrame import DirectFrame
from . import DirectGuiHelper as DGH


class DirectItemContainer():
    def __init__(self, element, rowIndex, columnIndex, widthInColumns, heightInRows):
        self.element = element
        self.rowIndex = rowIndex
        self.columnIndex = columnIndex
        self.widthInColumns = widthInColumns
        self.heightInRows = heightInRows

class DirectGridSizer(DirectFrame):
    """
    A frame to add multiple other directGui elements to that will then be
    automatically be placed stacked next to each other.
    """
    def __init__(self, parent = None, **kw):
        optiondefs = (
            # Define type of DirectGuiWidget
            ('items',          [],          self.refresh),
            ('pgFunc',         PGItem,      None),
            ('numStates',      1,           None),
            ('state',          DGG.NORMAL,  None),
            ('borderWidth',    (0, 0),      self.setBorderWidth),

            ('itemMargin',     (0,0,0,0),   self.refresh),
            ('numRows',        4,           self.refresh),
            ('numColumns',     4,           self.refresh),
            ('autoUpdateFrameSize', True,   None),
            ('boxAlign',    TextNode.ALeft, self.refresh),

            ('suppressMouse',  0,           None),
            )
        # Merge keyword options with default options
        self.defineoptions(kw, optiondefs)

        # Initialize superclasses
        DirectFrame.__init__(self, parent)

        # Call option initialization functions
        self.initialiseoptions(DirectGridSizer)

    def addItem(self, element, row, column, widthInColumns=1, heightInRows=1):
        """
        Adds the given item to this panel stack
        """
        element.reparentTo(self)
        container = DirectItemContainer(element, row, column, widthInColumns, heightInRows)
        self["items"].append(container)
        self.refresh()

    def removeItem(self, element):
        """
        Remove this item from the panel
        """
        for item in self["items"]:
            if element == item.element:
                self["items"].remove(item)
                self.refresh()
                return 1
        return 0

    def clearItems(self):
        for item in self["items"][:]:
            self["items"].remove(item)
        self.refresh()

    def refresh(self):
        """
        Recalculate the position of every item in this panel and set the frame-
        size of the panel accordingly if auto update is enabled.
        """
        # sanity check so we don't get here to early

        if not hasattr(self, "bounds"): return

        for item in self["items"]:
            item.element.frameInitialiseFunc()

        # variables to store the maximum row and column widths
        rowHeights = [0]*self["numRows"]
        columnWidths = [0]*self["numColumns"]

        b_top = 0
        b_bottom = 0
        b_left = 0
        b_right = 0

        pad_x = self["pad"][0]
        pad_y = self["pad"][1]

        margin_left = self["itemMargin"][0]
        margin_right = self["itemMargin"][1]
        margin_bottom = self["itemMargin"][2]
        margin_top = self["itemMargin"][3]

        if self["autoUpdateFrameSize"]:
            self["frameSize"] = (0, 0, 0, 0)

        # get the max row and column sizes
        for r in range(self["numRows"]):
            for c in range(self["numColumns"]):

                for item in self["items"]:
                    if item.rowIndex == r and item.columnIndex == c:
                        rowHeights[r] = max(rowHeights[r], DGH.getRealHeight(item.element) / item.heightInRows + margin_bottom + margin_top)
                        columnWidths[c] = max(columnWidths[c], DGH.getRealWidth(item.element) / item.widthInColumns + margin_left + margin_right)
                        break

        for item in self["items"]:
            r = item.rowIndex
            c = item.columnIndex

            if r >= self["numRows"]:
                raise IndexError(f"Row index defined in item {item.element} exceeded number of rows in grid sizer: numRows={self['numRows']}")
            if c >= self["numColumns"]:
                raise IndexError(f"Column index defined in item {item.element} exceeded number of columns in grid sizer: numColumns={self['numColumns']}")

            z = 0
            for i in range(r):
                z -= rowHeights[i]

            x = 0
            for i in range(c):
                x += columnWidths[i]

            item.element.setPos(pad_x + x + margin_left, 0, pad_y + z + margin_top)

        if self["autoUpdateFrameSize"]:

            for item in self["items"]:
                b_left = min(b_left, DGH.getRealLeft(item.element) + item.element.getX())
                b_right = max(b_right, DGH.getRealRight(item.element) + item.element.getX())

                b_bottom = min(b_bottom, DGH.getRealBottom(item.element) + item.element.getZ())
                b_top = max(b_top, DGH.getRealTop(item.element) + item.element.getZ())

            self["frameSize"] = [b_left+pad_x, b_right+pad_x, b_bottom+pad_y, b_top+pad_y]

            xShift = 0
            if self["boxAlign"] == TextNode.ACenter:
                self["frameSize"][0] = self["frameSize"][0] - self["frameSize"][1]/2
                self["frameSize"][1] = self["frameSize"][1]/2
                xShift = -self["frameSize"][1]/2
            if self["boxAlign"] == TextNode.ARight:
                self["frameSize"][0] = self["frameSize"][1]
                self["frameSize"][1] = self["frameSize"][0]
                xShift = -self["frameSize"][1]



'''This module contains the DirectSpinBox class.

A DirectSpinBox is a type of Entry specialized for number selection.
It contains two buttons to its right for increasing and decreasing the
numeric value given in the entry field.  Values can also be directly
typed in the entry and will be checked against the type given to the
spiner box which is *int* by default.

Options available for this Widget are:

*   value
    The value selected by the spiner. This will always return the number
    in the format given in the valueType option.  You can also pass a
    string or other type which the valueType function is possible to
    convert from.

*   textFormat
    This is a format string which will be used to write the value into
    the entry field

*   stepSize
    When the up and down arrows are pressed, stepSize will be used to
    determine how much the value should be in-/decreased

*   minValue
    The minimum value that can be entered in the spiner value. If values
    given to the spiner are lower than the min value, the spiner will
    set the minValue as new value.

*   maxValue
    The maximum value that can be entered in the spiner value. If values
    given to the spiner are larger than the max value, the spiner will
    set the maxValue as new value.

*   valueType
    A function like float, int or similar which will convert the value
    given to the spiner.  It's return value if it was able to convert
    the value correctly will then be converted using the textFormat to
    a string which is entered into the textfield.

*   repeatdelay
    Delay in seconds at which the value should be in-/decreased when the
    up or down arrow button is held down

*   repeatStartdelay
    This delay will be used to determine when the repeat functionality
    is actually started after the user presses down any of the buttons

*   command
    This command will be passed to the entry field and hence will be
    called, whenever the user presses enter in it.

*   extraArgs
    Extra arguments passed to the entry field for when the command is
    used

*   incButtonCallback
    A callback function which will be called when the increase button is
    first pressed.  It won't be called when the button is held down and
    the value is changed repeatedly.

*   decButtonCallback
    A callback function which will be called when the decrease button is
    first pressed.  It won't be called when the button is held down and
    the value is changed repeatedly.

'''

__all__ = ['DirectSpinBox']

from panda3d.core import *
from direct.gui import DirectGuiGlobals as DGG
from direct.directnotify import DirectNotifyGlobal
from direct.task.Task import Task
from direct.gui.DirectFrame import *
from direct.gui.DirectEntry import *
from direct.gui.DirectButton import *

from direct.showbase.MessengerGlobal import messenger
from direct.task.TaskManagerGlobal import taskMgr

DGG.MWUP = PGButton.getPressPrefix() + MouseButton.wheel_up().getName() + '-'
DGG.MWDOWN = PGButton.getPressPrefix() + MouseButton.wheel_down().getName() + '-'

class DirectSpinBox(DirectFrame):
    notify = DirectNotifyGlobal.directNotify.newCategory('DirectSpinBox')

    def __init__(self, parent = None, **kw):
        assert self.notify.debugStateCall(self)

        # Inherits from DirectFrame
        optiondefs = (
            # Define type of DirectGuiWidget
            ('value',              0,         None),
            ('textFormat',         '{:0d}',   None),
            ('stepSize',           1,         None),
            ('minValue',           0,         None),
            ('maxValue',           100,       None),
            ('valueType',          int,       None),
            ('repeatdelay',        0.125,     None),
            ('repeatStartdelay',   0.25,      None),
            ('command',            None,      None),
            ('extraArgs',          [],        None),
            ('incButtonCallback',  None,      self.setIncButtonCallback),
            ('decButtonCallback',  None,      self.setDecButtonCallback),
            ('buttonOrientation',  DGG.VERTICAL, None),
            ('state',          DGG.NORMAL, None),
            # set default border width to 0
            ('borderWidth',        (0,0),     None),
            )
        # Merge keyword options with default options
        self.defineoptions(kw, optiondefs)

        # Initialize superclasses
        DirectFrame.__init__(self, parent)

        # create the textfield that will hold the text value
        self.valueEntry = self.createcomponent(
            'valueEntry', (), None,
            DirectEntry, (self,),
            #overflow = 1, #DOESN'T WORK WITH RIGHT TEXT ALIGN!!!
            width = 2,
            text_align = TextNode.ARight,
            command = self['command'],
            extraArgs = self['extraArgs'],
            focusOutCommand = self.focusOutCommand,
            )
        self.valueEntry.bind(DGG.MWUP, self.__mousewheelUp)
        self.valueEntry.bind(DGG.MWDOWN, self.__mousewheelDown)

        # try set the initial value
        try:
            self.setValue(self['value'])
        except:
            # Make sure the initial value is good
            self.setValue(0)

        # This font contains the up and down arrow
        shuttle_controls_font = loader.loadFont('shuttle_controls')

        # create the up arrow button
        self.incButton = self.createcomponent(
            'incButton', (), None,
            DirectButton, (self,),
            text = '5' if self['buttonOrientation'] == DGG.VERTICAL else '4',
            text_font = shuttle_controls_font,
            )
        # Set commands for the Inc Button
        self.incButton.bind(DGG.B1PRESS, self.__incButtonDown)
        self.incButton.bind(DGG.B1RELEASE, self.__buttonUp)

        # create the down arrow button
        self.decButton = self.createcomponent(
            'decButton', (), None,
            DirectButton, (self,),
            text = '6' if self['buttonOrientation'] == DGG.VERTICAL else '3',
            text_font = shuttle_controls_font
            )
        # Set commands for the Dec Button
        self.decButton.bind(DGG.B1PRESS, self.__decButtonDown)
        self.decButton.bind(DGG.B1RELEASE, self.__buttonUp)

        # Set the spiners elements position
        self.resetPosition()
        if self['frameSize'] is None:
            # Calculate the spiners frame size only if we don't have a
            # custom frameSize
            self.recalcFrameSize()

        self.initialiseoptions(DirectSpinBox)

        # Here we check for custom values of properties.
        # We need to do this for all components which have been edited
        # by code after they have been set up by createcomponent.
        if self['incButton_pos'] is not None:
            self.incButton.setPos(self['incButton_pos'])
        if self['decButton_pos'] is not None:
            self.decButton.setPos(self['decButton_pos'])
        if self['valueEntry_pos'] is not None:
            self.valueEntry.setPos(self['valueEntry_pos'])


    def resetPosition(self):
        '''
        Positions the two buttons to the right of the text entry
        '''
        assert self.notify.debugStateCall(self)
        # Position the text Entry field centered vertically
        valCenter = self.valueEntry.getCenter()
        self.valueEntry.setPos(0, 0, -valCenter[1])

        # Position the Inc Button
        incCenter = self.incButton.getCenter()
        incWidth = self.incButton.getWidth()
        incHeight = self.incButton.getHeight()
        incBorderW = self.incButton['borderWidth'][0]
        incBorderH = self.incButton['borderWidth'][1]
        if self['buttonOrientation'] == DGG.VERTICAL:
            self.incButton.setPos(-incCenter[0]+incWidth/2+incBorderW, 0, -incCenter[1]+incHeight/2+incBorderH)
            self.incButton.setX(self.incButton, self.valueEntry.bounds[1])
        else:
            self.incButton.setPos(-incCenter[0]+incWidth/2+incBorderW, 0, -incCenter[1])
            self.incButton.setX(self.incButton, self.valueEntry.bounds[1])

        # Position the Dec Button
        decCenter = self.decButton.getCenter()
        decWidth = self.decButton.getWidth()
        decHeight = self.decButton.getHeight()
        decBorderW = self.decButton['borderWidth'][0]
        decBorderH = self.decButton['borderWidth'][1]
        if self['buttonOrientation'] == DGG.VERTICAL:
            self.decButton.setPos(-decCenter[0]+decWidth/2+decBorderW, 0, -decCenter[1]-decHeight/2-decBorderH)
            self.decButton.setX(self.decButton, self.valueEntry.bounds[1])
        else:
            valWidth = self.valueEntry.getWidth()
            self.decButton.setPos(-valWidth-decCenter[0]-decWidth/2-decBorderW, 0, -decCenter[1])
            self.decButton.setX(self.decButton, self.valueEntry.bounds[1])


    def recalcFrameSize(self):
        '''
        Set the surrounding frame so the spinner will actually look
        like a box and will be able to give correct values for
        functions and properties like getWidth, getCenter and bounds
        '''
        assert self.notify.debugStateCall(self)
        if self['buttonOrientation'] == DGG.VERTICAL:
            l = self.valueEntry.bounds[0] - self['borderWidth'][0]
            r = self.decButton.getX() + self.decButton.bounds[1] + self.decButton['borderWidth'][0] + self['borderWidth'][0]
            b = self.valueEntry.getZ() + self.valueEntry.bounds[2] - self['borderWidth'][1]
            t = self.valueEntry.getZ() + self.valueEntry.bounds[3] + self['borderWidth'][1]
        else:
            l = self.decButton.getX() + self.decButton.bounds[0] - self.incButton['borderWidth'][0] - self['borderWidth'][0]
            r = self.incButton.getX() + self.incButton.bounds[1] + self.incButton['borderWidth'][0] + self['borderWidth'][0]
            b = self.valueEntry.getZ() + self.valueEntry.bounds[2] - self['borderWidth'][1]
            t = self.valueEntry.getZ() + self.valueEntry.bounds[3] + self['borderWidth'][1]
        self['frameSize'] = (l, r, b, t)

    def __repeatStepTask(self, task):
        assert self.notify.debugStateCall(self)
        ret = self.doStep(task.stepSize)
        task.setDelay(self['repeatdelay'])
        if ret:
            return Task.again
        else:
            return Task.done

    def __incButtonDown(self, event):
        assert self.notify.debugStateCall(self)
        task = Task(self.__repeatStepTask)
        task.stepSize = self['stepSize']
        taskName = self.taskName('repeatStep')
        #print 'incButtonDown: adding ', taskName
        taskMgr.doMethodLater(self['repeatStartdelay'], task, taskName)
        self.doStep(task.stepSize)
        messenger.send('wakeup')
        if self.__incButtonCallback:
            self.__incButtonCallback()

    def __decButtonDown(self, event):
        assert self.notify.debugStateCall(self)
        task = Task(self.__repeatStepTask)
        task.stepSize = -self['stepSize']
        taskName = self.taskName('repeatStep')
        #print 'decButtonDown: adding ', taskName
        taskMgr.doMethodLater(self['repeatStartdelay'], task, taskName)
        self.doStep(task.stepSize)
        messenger.send('wakeup')
        if self.__decButtonCallback:
            self.__decButtonCallback()

    def __buttonUp(self, event):
        assert self.notify.debugStateCall(self)
        taskName = self.taskName('repeatStep')
        #print 'buttonUp: removing ', taskName
        taskMgr.remove(taskName)

    def __mousewheelUp(self, event):
        assert self.notify.debugStateCall(self)
        self.doStep(self['stepSize'])

    def __mousewheelDown(self, event):
        assert self.notify.debugStateCall(self)
        self.doStep(-self['stepSize'])

    def doStep(self, stepSize):
        """Adds the value given in stepSize to the current value stored
        in the spinner.  Pass a negative value to subtract from the
        current value.
        """
        assert self.notify.debugStateCall(self)
        #print 'doStep[', stepSize,']'

        return self.setValue(self['value'] + stepSize)

    def __checkValue(self, newValue):
        assert self.notify.debugStateCall(self)
        try:
            value = self['valueType'](newValue)
            if value < self['minValue']:
                self.notify.info('Value out of range value: {} min allowed value: {}'.format(value, self['minValue']))
                return self['minValue']
            if value > self['maxValue']:
                self.notify.info('Value out of range value: {} max allowed value: {}'.format(value, self['maxValue']))
                return self['maxValue']
            return value
        except:
            self.notify.info('ERROR: NAN {}'.format(newValue))
            return None

    def get(self):
        '''
        Returns the value in string format (see getValue to get the value in it's specific type)
        '''
        assert self.notify.debugStateCall(self)
        return self.valueEntry.get()

    def getValue(self):
        '''
        Returns the value in it's actual type (see get to get the value as a string)
        '''
        assert self.notify.debugStateCall(self)
        return self['value']

    def setValue(self, newValue):
        '''
        Set a new value for the spinbox to display. newValue can be any type which
        can be converted by the function set in valueType
        '''
        assert self.notify.debugStateCall(self)
        value = self.__checkValue(newValue)
        if value is None:
            self.valueEntry.enterText(self['textFormat'].format(self['value']))
            return False

        self.valueEntry.enterText(self['textFormat'].format(value))
        self['value'] = value
        return True

    def focusOutCommand(self):
        assert self.notify.debugStateCall(self)
        self.setValue(self.get())

    def setIncButtonCallback(self):
        assert self.notify.debugStateCall(self)
        self.__incButtonCallback = self['incButtonCallback']

    def setDecButtonCallback(self):
        assert self.notify.debugStateCall(self)
        self.__decButtonCallback = self['decButtonCallback']

'''
from direct.showbase.ShowBase import ShowBase
base = ShowBase()

spinBox = DirectSpinBox(pos=(0,0,-0.25), value=5, minValue=-100, maxValue=100, repeatdelay=0.125, buttonOrientation=DGG.HORIZONTAL, valueEntry_text_align=TextNode.ACenter, borderWidth=(1,1))
spinBox.setScale(0.1)
spinBox["relief"] = 2
spinBox = DirectSpinBox(pos=(0,0,0.25), valueEntry_width=10, borderWidth=(2,2), frameColor=(1,0,0,1))
spinBox.setScale(0.1)
base.run()
'''



"""This snippet shows how to create a tooltip text that will be attached
to the cursor and check it's position to not move out of the screen."""

__all__ = ['DirectTooltip']

import sys

from panda3d.core import TextNode
from direct.gui.DirectGui import DirectLabel

class DirectTooltip(DirectLabel):
    def __init__(self, parent = None, **kw):
        optiondefs = (
            ('text',        'Tooltip',      None),
            ('text_align',  TextNode.ALeft, None),
            #('text_fg',    (1, 1, 1, 1),    None),
            #('text_bg',    (0, 0, 0, 0.75), None),
            #('text_frame', (0, 0, 0, 0.75), None),
            ('borderWidth', (0.05, 0.05),   None),
            #('parent',      base.pixel2d,   None),
            #('sortOrder',   1000,           None),
           )
        # Merge keyword options with default options
        self.defineoptions(kw, optiondefs)
        # Initialize superclasses
        DirectLabel.__init__(self, parent)

        # make sure the text apears slightly right below the cursor
        if parent is base.pixel2d:
            self.textXShift = 10
            self.textYShift = -32
        else:
            self.textXShift = 0.05
            self.textYShift = -0.08

        self.mousePos = None

        # this will determine when the tooltip should be moved in the
        # respective direction, whereby
        # 1  : display edge
        # <1 : margin inside the window
        # >1 : margin outside the window
        self.xEdgeStartShift = 0.99
        self.yEdgeStartShift = 0.99

        # make sure the tooltip is shown on top
        self.setBin('gui-popup', 0)

        # Call option initialization functions
        self.initialiseoptions(DirectTooltip)

        self.hide()

    def show(self, text=None, args=None):
        if text is not None:
            self.setText(text)
            self.resetFrameSize()#setFrameSize(True)
        DirectLabel.show(self)

        # add the tooltips update task so it will be updated every frame
        base.taskMgr.add(self.updateTooltipPos, self.taskName("task_updateTooltipPos"))

    def hide(self, args=None):
        DirectLabel.hide(self)

        # remove the tooltips update task
        base.taskMgr.remove(self.taskName("task_updateTooltipPos"))

    def updateTooltipPos(self, task):
        # calculate new aspec tratio
        wp = base.win.getProperties()
        aspX = 1.0
        aspY = 1.0
        wpXSize = wp.getXSize()
        wpYSize = wp.getYSize()
        if self.getParent() != base.pixel2d:
            # calculate the aspect ratio of the window if we're not reparented
            # to the pixel2d nodepath
            if wpXSize > wpYSize:
                aspX = wpXSize / float(wpYSize)
            else:
                aspY = wpYSize / float(wpXSize)

        # variables to store the mouses current x and y position
        x = 0.0
        y = 0.0
        if base.mouseWatcherNode.hasMouse():
            DirectLabel.show(self)

            # Move the tooltip to the mouse

            # get the mouse position
            if self.getParent() == base.pixel2d:
                x = base.win.getPointer(0).getX()
                y = -base.win.getPointer(0).getY()
            else:
                x = (base.mouseWatcherNode.getMouseX()*aspX)
                y = (base.mouseWatcherNode.getMouseY()*aspY)

            # set the text to the current mouse position
            self.setPos(
                x + self.textXShift,
                0,
                y + self.textYShift)

            bounds = self.getBounds()
            # bounds = left, right, bottom, top

            # calculate the texts bounds respecting its current position
            xLeft = self.getX() + bounds[0]*self.getScale()[0]
            xRight = self.getX() + bounds[1]*self.getScale()[0]
            yUp = self.getZ() + bounds[3]*self.getScale()[1]
            yDown = self.getZ() + bounds[2]*self.getScale()[1]

            # these will be used to shift the text in the desired direction
            xShift = 0.0
            yShift = 0.0

            if self.getParent() == base.pixel2d:
                # make sure to have the correct edges in a pixel2d environment
                # even if the window has been resized
                self.xEdgeStartShift = wpXSize * 0.99
                self.yEdgeStartShift = wpYSize * 0.99

            if xRight/aspX > self.xEdgeStartShift:
                # shift to the left
                xShift = self.xEdgeStartShift - xRight/aspX
            elif xLeft/aspX < -self.xEdgeStartShift:
                # shift to the right
                xShift = -(self.xEdgeStartShift + xLeft/aspX)
            if yUp/aspY > self.yEdgeStartShift:
                # shift down
                yShift = self.yEdgeStartShift - yUp/aspY
            elif yDown/aspY < -self.yEdgeStartShift:
                # shift up
                yShift = -(self.yEdgeStartShift + yDown/aspY)

            # some aspect ratio calculation
            xShift *= aspX
            yShift *= aspY

            # move the tooltip to the new position
            self.setX(self.getX() + xShift)
            self.setZ(self.getZ() + yShift)

        else:
            DirectLabel.hide(self)


        # continue the task until it got manually stopped
        return task.cont




"""This module contains the DirectBoxSizer class."""

__all__ = ['DirectGridSizer']

import calendar
import datetime
from panda3d.core import *
from direct.gui import DirectGuiGlobals as DGG
from direct.gui.DirectButton import DirectButton
from direct.gui.DirectLabel import DirectLabel
from . import DirectGuiHelper as DGH
from .DirectGridSizer import DirectGridSizer
from .DirectSpinBox import DirectSpinBox
from .DirectOptionMenu import DirectOptionMenu


class DirectDatePicker(DirectGridSizer):
    def __init__(self, parent = None, **kw):
        optiondefs = (
            ('year',    None,   self.refreshPicker),
            ('month',   None,   self.refreshPicker),
            ('day',     None,   self.refreshPicker),

            ('normalDayFrameColor', None, None),
            ('activeDayFrameColor', None, None),
            ('todayFrameColor', None, None),

            # Define type of DirectGuiWidget
            )
        # Merge keyword options with default options
        self.defineoptions(kw, optiondefs)

        # Initialize superclasses
        DirectGridSizer.__init__(self, parent, numRows=7, numColumns=8, autoUpdateFrameSize=True, itemMargin=[0.005, 0.005, 0.005, 0.005])

        # Call option initialization functions
        self.initialiseoptions(DirectDatePicker)

        # set up the calendar
        self.cal = calendar.Calendar()

        # normal day colors
        self['normalDayFrameColor'] = (
            (0.7,0.7,0.7,1),
            (0.7,0.7,0.7,1),
            (0.7,0.0,0.0,1),
            (0.6,0.6,0.6,1)
        ) if self['normalDayFrameColor'] is None else self['normalDayFrameColor']

        # selected day color
        self['activeDayFrameColor'] = (
            (0.8,0.8,0.8,1),
            (0.8,0.8,0.8,1),
            (0.8,0.0,0.0,1),
            (0.6,0.6,0.6,1)
        ) if self['activeDayFrameColor'] is None else self['activeDayFrameColor']

        # the systems current day color
        self['todayFrameColor'] = (
            (1.0,0.7,0.7,1),
            (0.8,0.8,0.8,1),
            (0.8,0.0,0.0,1),
            (0.6,0.6,0.6,1)
        ) if self['todayFrameColor'] is None else self['todayFrameColor']

        # current date
        now = datetime.datetime.now()

        # add the week header
        weekDays = calendar.weekheader(2).split(" ")
        self.headers = []
        for i in range(1, 8):
            lbl = self.createLabel(str(weekDays[i-1]))
            self.headers.append(lbl)
            self.addItem(lbl, 0, i)

        # add dummy week numbers
        self.weekNumbers = []
        for i in range(1, 7):
            lbl = self.createLabel("--")
            self.weekNumbers.append(lbl)
            self.addItem(lbl, i, 0)

        self.dateButtons = []
        day = 0
        for row in range(1, 7):
            self.dateButtons.append([])
            for column in range(1, 8):
                btn = self.createDateButton(day, False)
                self.dateButtons[row-1].append(btn)
                self.addItem(btn, row, column)
                day += 1

        self.ll = Point3(self["frameSize"][0], self["frameSize"][2])
        self.ur = Point3(self["frameSize"][1], self["frameSize"][3])

        # the year picker
        self.yearPicker = self.createcomponent(
            'yearPicker', (), None,
            DirectSpinBox, (self,),
            valueEntry_width=5,
            scale=0.05,
            minValue=1,
            maxValue=9999,
            repeatdelay=0.125,
            buttonOrientation=DGG.HORIZONTAL,
            valueEntry_text_align=TextNode.ACenter,
            borderWidth=(.1,.1),
            incButtonCallback=self.__changedYear,
            decButtonCallback=self.__changedYear,
            command=self.__changedYear)
        self.yearPicker.resetFrameSize()
        self.yearPicker.setPos(
            DGH.getRealLeft(self) + DGH.getRealWidth(self.yearPicker)/2,
            0,
            DGH.getRealTop(self) - DGH.getRealBottom(self.yearPicker))

        self.monthPicker = self.createcomponent(
            'monthPicker', (), None,
            DirectOptionMenu, (self,),
            items=calendar.month_name[1:],
            scale=0.05,
            command=self.setMonth,
            relief=DGG.FLAT)
        self.monthPicker.resetFrameSize()

        self.monthPicker.setPos(
            DGH.getRealRight(self) - DGH.getRealWidth(self.monthPicker),
            0,
            DGH.getRealTop(self) - DGH.getRealBottom(self.monthPicker) / 2)


        # set the current date if none is given at initialization
        self['year'] = now.year if self['year'] is None else self['year']
        self['month'] = now.month if self['month'] is None else self['month']
        self['day'] = now.day if self['day'] is None else self['day']

        self.yearPicker.setValue(self['year'])
        self.monthPicker.set(self['month']-1)

        self.refreshPicker()

    def createDateButton(self, day, enabled=True):
        return self.createcomponent(
            'day{}'.format(day), (), None,
            DirectButton, (self,),
            text=str(day),
            text_scale=0.05,
            relief=DGG.FLAT,
            borderWidth=(0.01, 0.01),
            state=DGG.NORMAL if enabled else DGG.DISABLED,
            frameColor=self['normalDayFrameColor'],
            frameSize=(-0.05,0.05,-0.035,0.0625),
            command=self.setDay,
            extraArgs=[day])

    def createLabel(self, txt):
        return self.createcomponent(
            'weekNumber', (), None,
            DirectLabel, (self,),
            text=txt,
            text_scale=0.05,
            frameColor=(0,0,0,0),
            frameSize=(-0.05,0.05,-0.035,0.0625)
        )

    def setDay(self, day):
        self['day'] = day
        self.refreshPicker()

    def getDay(self):
        return self['day']

    def __changedYear(self, year=None):
        year = int(self.yearPicker.get()) if year is None else int(year)
        if year > 9999: year = 9999
        if year < 1: year = 1
        self.setYear(year)

    def setMonth(self, month):
        if type(month) == str:
            self['month'] = calendar.month_name[:].index(month)
            if self.monthPicker.get() != self['month']-1:
                self.monthPicker.set(self['month']-1, fCommand=False)
        else:
            self['month'] = month
            if self.monthPicker.get() != self['month']-1:
                self.monthPicker.set(self['month']-1, fCommand=False)
        self.refreshPicker

    def getMonth(self):
        return self['month']

    def setYear(self, year):
        self['year'] = year
        self.refreshPicker()

    def getYear(self):
        return self['year']

    def getDate(self):
        return datetime.datetime(self['year'], self['month'], self['day'])

    def get(self):
        return self.getDate()

    def refreshPicker(self):
        # sanity check so we don't get here to early
        if self['year'] is None or self['month'] is None or self['day'] is None:
            return

        datesPrev = None
        if calendar.monthrange(self['year'], self['month'])[0] == 0:
            # get the previous dates for months that start with monday as first weekday and need one week prepended
            year = self['year'] if self['month'] > 1 else self['year']-1 if self['year'] > 0 else 1
            month = self['month']-1 if self['month'] > 1 else 12
            datesPrev = self.cal.monthdatescalendar(year,month)

        dates = self.cal.monthdatescalendar(self['year'],self['month'])

        datesNext = None
        if datesPrev is None and len(dates) < 7:
            # get the next dates for months that do not cover 6 lines of the calendar
            year = self['year'] if self['month'] < 12 else self['year']+1 if self['year'] < 9999 else 9999
            month = self['month']+1 if self['month'] < 12 else 1
            datesNext = self.cal.monthdatescalendar(year,month)


        def updateInfo(row, column, date):
            # update button
            btn = self.dateButtons[row][column]
            btn["text"] = str(date.day)
            btn["extraArgs"] = [date.day]
            btn["state"] = DGG.NORMAL if date.month == self['month'] else DGG.DISABLED
            now = datetime.datetime.now()
            isToday = now.day == date.day and now.month == date.month and now.year == date.year
            if isToday:
                # the current day
                btn["frameColor"] = self['todayFrameColor']
            if self['day'] == date.day:
                # the selected one
                btn["frameColor"] = self['activeDayFrameColor']
            elif not isToday:
                btn["frameColor"] = self['normalDayFrameColor']

            # update week numbers
            if column == 0:
                weeknumber = datetime.date(date.year, date.month, date.day).isocalendar()[1]
                self.weekNumbers[row]["text"] = str(weeknumber)

        for row in range(1, 7):
            r = row - 1
            column = 0
            if datesPrev and r == 0:
                for date in datesPrev[-1]:
                    updateInfo(r, column, date)
                    column += 1

            elif len(dates) > r - (1 if datesPrev else 0):
                hadPrev = 1 if datesPrev else 0
                for date in dates[r - hadPrev]:
                    updateInfo(r, column, date)
                    column += 1

            elif datesNext:
                idx = 1
                if dates[-1][-1].month == self['month']:
                    idx = 0

                for date in datesNext[idx]:
                    updateInfo(r, column, date)
                    column += 1




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="DirectGuiExtension",
    version="21.05",
    author="Fireclaw",
    author_email="fireclawthefox@gmail.com",
    description="A set of extensions for the DirectGUI system of the Panda3D engine",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fireclawthefox/DirectGuiExtension",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Software Development :: User Interfaces",
    ],
    install_requires=[
        'panda3d',
    ],
    python_requires='>=3.6',
)
